#include "ros/ros.h"
#include <octomap_msgs/Octomap.h>
#include <octomap_msgs/conversions.h>
#include <octomap/OcTree.h>
#include <cstring>
#include <iostream>
#include <memory>
#include <fstream>
#include <ostream>
#include <boost/program_options.hpp>
#include <ompl/base/Planner.h>
#include <ompl/base/ScopedState.h>
#include <ompl/base/spaces/DubinsAirplaneStateSpace.h>
#include <ompl/geometric/SimpleSetup.h>
#include <ompl/geometric/planners/bitstar/BITstar.h>
#include <ompl/geometric/planners/prm/PRMstar.h>
#include <ompl/geometric/planners/fmt/BFMT.h>
#include <ompl/geometric/planners/fmt/FMT.h>
#include <ompl/geometric/planners/prm/SPARS.h>
#include <ompl/geometric/planners/rrt/InformedRRTstar.h>
#include <fcl/geometry/octree/octree.h>
#include <fcl/geometry/shape/box.h>
#include <fcl/geometry/shape/cylinder.h>
#include <fcl/geometry/shape/sphere.h>
#include <fcl/narrowphase/collision.h>
#include <fcl/broadphase/broadphase_dynamic_AABB_tree.h>

#include "ompl_planner_utils.h"

namespace ob = ompl::base;
namespace og = ompl::geometric;
namespace po = boost::program_options;

int n = 0;
int k = 0;
double maxHeigth = -1.0;
double totalRunTime = 0.0;
double numSamples = 1000;
double interpolation = 0.05;
double solverRunTime = 10;
double turnRadius = 10;
double climbAngle = 0.15;
double volumeEdge = 20;
std::string plannerType = "KPIECE";
std::stringstream outputFilePath;

bool isStateValid(const ob::SpaceInformation *si, const ob::State *state, std::shared_ptr<octomap::OcTree> map, std::shared_ptr<fcl::DynamicAABBTreeCollisionManagerd> manager)
{
    const auto s = state->as<ob::SimpleSE3StateSpace::StateType>();

    /* auto n = map->search(s->getX(), s->getY(), s->getZ()); */
    /* if(n == NULL) */
    /*     return false; */

    /* if(map->isNodeOccupied(n)) */
    /*     return false; */
	if(maxHeigth > 0.0 && s->getZ() > maxHeigth)
	{
		return false;
	}

    std::shared_ptr<fcl::Sphered> geom(new fcl::Sphered(volumeEdge));
    fcl::Transform3d pose = fcl::Transform3d::Identity();
    fcl::Vector3d T = fcl::Vector3d(s->getX(), s->getY(), s->getZ());
    pose.translation() = T;
    fcl::CollisionObjectd* obj = new fcl::CollisionObjectd(geom, pose);
    CollisionData<double> cdata;
    manager->collide(obj, &cdata, defaultCollisionFunction);

    delete obj;

    return si->satisfiesBounds(state) && !cdata.result.isCollision();
}

void plan(DummyState _start, DummyState _goal, std::shared_ptr<octomap::OcTree> map, std::shared_ptr<fcl::DynamicAABBTreeCollisionManagerd> manager, std::vector<og::PathGeometric> &paths)
{
    ob::StateSpacePtr space = std::make_shared<ob::DubinsAirplaneStateSpace>(turnRadius, climbAngle, false);

    double xMax, yMax, zMax;
    map->getMetricMax(xMax, yMax, zMax);
    double xMin, yMin, zMin;
    map->getMetricMin(xMin, yMin, zMin);

    ob::RealVectorBounds bounds(3);
    bounds.setHigh(0, xMax);
    bounds.setHigh(1, yMax);
    bounds.setHigh(2, zMax);

    bounds.setLow(0, xMin);
    bounds.setLow(1, yMin);
    bounds.setLow(2, zMin);

    space->as<ob::SimpleSE3StateSpace>()->setBounds(bounds);

    ob::ScopedState<ob::SimpleSE3StateSpace> start(space);
    start->setXYZYaw(_start.x, _start.y, _start.z, _start.yaw);
    ob::ScopedState<ob::SimpleSE3StateSpace> goal(space);
    goal->setXYZYaw(_goal.x, _goal.y, _goal.z, _goal.yaw);

    og::SimpleSetup ss(space);
    ob::SpaceInformation *si = ss.getSpaceInformation().get();
    ss.setStateValidityChecker([si, map, manager](const ob::State *state) {
        return isStateValid(si, state, map, manager);
    });
    si->setMotionValidator(std::make_shared<ob::DubinsAirplaneMotionValidator>(si));
	si->setStateValidityCheckingResolution(0.001);
	auto res = si->getStateValidityCheckingResolution();
	ROS_INFO_STREAM("Validity res: " << res);

    ss.setStartAndGoalStates(start, goal);

    if(plannerType == "FMT")
    {
        auto planner(std::make_shared<og::FMT>(ss.getSpaceInformation()));
		planner->setNumSamples((int) numSamples);
        ss.setPlanner(planner);
    }
    else if(plannerType == "RRTstar")
    {
        auto planner(std::make_shared<og::RRTstar>(ss.getSpaceInformation()));
        ss.setPlanner(planner);
    }
    else if(plannerType == "kBITstar")
    {
        auto planner(std::make_shared<og::BITstar>(ss.getSpaceInformation()));
		planner->setSamplesPerBatch(k);
		/* planner->setDelayRewiringUntilInitialSolution(true); */
        ss.setPlanner(planner);
    }
    else if(plannerType == "rBITstar")
    {
        auto planner(std::make_shared<og::BITstar>(ss.getSpaceInformation()));
		planner->setSamplesPerBatch(k);
		planner->setDelayRewiringUntilInitialSolution(true);
		planner->setUseKNearest(false);
        ss.setPlanner(planner);
    }
    else if(plannerType == "InformedRRTstar")
    {
        auto planner(std::make_shared<og::InformedRRTstar>(ss.getSpaceInformation()));
        ss.setPlanner(planner);
    }
    else if(plannerType == "SPARS")
    {
        auto planner(std::make_shared<og::SPARS>(ss.getSpaceInformation()));
        ss.setPlanner(planner);
    }
    else if(plannerType == "BFMT")
    {
        auto planner(std::make_shared<og::BFMT>(ss.getSpaceInformation()));
		planner->setNumSamples((int) numSamples);
        ss.setPlanner(planner);
    }
    else if(plannerType == "PRMstar")
    {
        auto planner(std::make_shared<og::PRMstar>(ss.getSpaceInformation()));
        ss.setPlanner(planner);
    }

    ss.setup();

    ob::PlannerStatus solved = ss.solve(solverRunTime);

    if (solved)
    {
		totalRunTime += ss.getLastPlanComputationTime();
        ss.simplifySolution();
        og::PathGeometric path = ss.getSolutionPath();
        paths.push_back(path);
    }
    else
        ROS_ERROR("NO SOLUTION FOUND");
        return;

    ss.clear();
    return;
}


void writeFail()
{
    std::ofstream pathfile;
    pathfile.open(outputFilePath.str());
    pathfile << "# " << n << "\tNO SOLUTION FOUND \n";
    pathfile.close();
}

void writePaths(std::vector<og::PathGeometric> &paths)
{
    double len = 0.0;
    std::stringstream path;
    for (int i = 0; i < paths.size(); i++)
    {
        len += paths[i].length();
        paths[i].interpolate(interpolation * ceil(paths[i].length()));
        paths[i].printAsMatrix(path);

    }
    std::ofstream pathfile;
    pathfile.open(outputFilePath.str());
    pathfile << n << "\t" << totalRunTime  << "\t" << std::setprecision(9) << len << "\n" << path.str();
    pathfile.close();
}

std::shared_ptr<octomap::OcTree> readOctomapTopic()
{
    auto msg = ros::topic::waitForMessage<octomap_msgs::Octomap>("octomap_full");
    auto tree = octomap_msgs::msgToMap(*msg);
    octomap::OcTree* ot = dynamic_cast<octomap::OcTree*>(tree);
    auto map = std::make_shared<octomap::OcTree>(*ot);
    return map;
}

std::shared_ptr<fcl::DynamicAABBTreeCollisionManagerd> setupCollisionManager(std::shared_ptr<octomap::OcTree> map, std::vector<NoFlyZone> zones)
{
    std::shared_ptr<fcl::DynamicAABBTreeCollisionManagerd> manager = std::make_shared<fcl::DynamicAABBTreeCollisionManagerd>();

    fcl::OcTree<double>* fclTree = new fcl::OcTree<double>(map);
    std::vector<fcl::CollisionObjectd*> objects;

    std::vector<std::array<double, 6> > boxes_ = fclTree->toBoxes();
    for(std::size_t i = 0; i < boxes_.size(); ++i)
    {
        double x = boxes_[i][0];
        double y = boxes_[i][1];
        double z = boxes_[i][2];
        double size = boxes_[i][3];
        double cost = boxes_[i][4];
        double threshold = boxes_[i][5];

        fcl::Box<double>* box = new fcl::Box<double>(size, size, size);
        box->cost_density = cost;
        box->threshold_occupied = threshold;
        fcl::CollisionObject<double>* obj = new fcl::CollisionObject<double>(
            std::shared_ptr<fcl::CollisionGeometry<double>>(box),
            fcl::Transform3<double>(fcl::Translation3<double>(fcl::Vector3<double>(x, y, z))));
        objects.push_back(obj);
    }

	for (std::size_t i = 0; i < zones.size(); ++i) {
		fcl::Cylinder<double>* cylinder = new fcl::Cylinder<double>(zones[i].radius, zones[i].heigth);
        fcl::CollisionObject<double>* obj = new fcl::CollisionObject<double>(
            std::shared_ptr<fcl::CollisionGeometry<double>>(cylinder),
            fcl::Transform3<double>(fcl::Translation3<double>(fcl::Vector3<double>(zones[i].x, zones[i].y, zones[i].heigth/2))));
        objects.push_back(obj);
	}

    manager->registerObjects(objects);
    manager->setup();

    return manager;
}

int parseArgs(int argc, char *argv[])
{
    for (int i = 1; i < argc; i++)
    {
        char *arg = argv[i];
        std::cout << arg << " ";
        if (strcmp(arg, "--planner") == 0 || strcmp(arg, "-p") == 0)
        {
            plannerType = argv[i + 1];
        }


        if (strcmp(arg, "--time") == 0 || strcmp(arg, "-t") == 0)
        {
            if (legalNumber(argv[i + 1]))
            {
                solverRunTime = atol(argv[i + 1]);
            }
        }

        if (strcmp(arg, "--number") == 0 || strcmp(arg, "-n") == 0)
        {
            if (legalNumber(argv[i + 1]))
            {
                n = atoi(argv[i + 1]);
            }
        }

        if (strcmp(arg, "--knumber") == 0 || strcmp(arg, "-k") == 0)
        {
            if (legalNumber(argv[i + 1]))
            {
                k = atoi(argv[i + 1]);
            }
        }

        if (strcmp(arg, "--outputFolder") == 0 || strcmp(arg, "-of") == 0)
        {
            outputFilePath << argv[i + 1];
        }

    }

    outputFilePath << plannerType << "_" << std::setfill('0') << std::setw(4) << solverRunTime << "s_" << std::setfill('0') << std::setw(3) << std::to_string(n) << ".txt";

	if (plannerType == "FMT" || plannerType == "BFMT")
	{
		numSamples = solverRunTime;
		solverRunTime = 600.0;
	}

    std::cout << "\n";
}

void readParams(ros::NodeHandle &n, std::vector<DummyState> &checkpoints, std::vector<NoFlyZone> &zones)
{
    if (n.getParam("/ompl_planner/turn_radius", turnRadius))
    {
        ROS_INFO_STREAM("Read turn radius from parameter server: " << turnRadius);
    }
    if (n.getParam("/ompl_planner/climb_angle", climbAngle))
    {
        ROS_INFO_STREAM("Read climb angle from parameter server: " << climbAngle);
    }
    if (n.getParam("/ompl_planner/volume_edge", volumeEdge))
    {
        ROS_INFO_STREAM("Read safe volume edge length from parameter server: " << volumeEdge);
    }
    if (n.getParam("/ompl_planner/max_heigth", maxHeigth))
    {
        ROS_INFO_STREAM("Read max heigth from parameter server: " << maxHeigth);
    }


    XmlRpc::XmlRpcValue cs; // checkpoint struct from parameter server
    if (n.getParam("/ompl_planner/checkpoints", cs))
    {
        ROS_ASSERT(cs.getType() == XmlRpc::XmlRpcValue::TypeArray);
        ROS_ASSERT_MSG(cs.size() > 1, "Not two or more checkpoints.");

        for (size_t i = 0; i < cs.size(); i++)
        {
            ROS_ASSERT(cs[i].getType() == XmlRpc::XmlRpcValue::TypeArray);
            ROS_ASSERT_MSG(cs[i].size() == 4, "Checkpoint does now have 4 values.");

            DummyState point(cs[i][0], cs[i][1], cs[i][2], cs[i][3]);

            checkpoints.push_back(point);
        }
        ROS_INFO_STREAM("Insterted " << cs.size() << " checkpoints.");
    }

    XmlRpc::XmlRpcValue nfzs; // noflyzone struct from parameter server
    if (n.getParam("/ompl_planner/noflyzones", nfzs))
    {
        ROS_ASSERT(nfzs.getType() == XmlRpc::XmlRpcValue::TypeArray);

        for (size_t i = 0; i < nfzs.size(); i++)
        {
            ROS_ASSERT(nfzs[i].getType() == XmlRpc::XmlRpcValue::TypeArray);
            ROS_ASSERT_MSG(nfzs[i].size() == 4, "Zone does now have 4 values.");

            NoFlyZone zone(nfzs[i][0], nfzs[i][1], nfzs[i][2], nfzs[i][3]);

            zones.push_back(zone);
        }
        ROS_INFO_STREAM("Insterted " << zones.size() << " NoFlyZones.");
    }
}

int main(int argc, char *argv[])
{
	/* ompl::msg::setLogLevel(ompl::msg::LOG_INFO); */
    ros::init(argc, argv, "ompl_planner", ros::init_options::AnonymousName);
    ros::NodeHandle n;

    std::vector<DummyState> checkpoints;
    std::vector<NoFlyZone> zones;
    std::vector<og::PathGeometric> paths;
    ob::StateSpacePtr space;

    readParams(n, checkpoints, zones);
    parseArgs(argc, argv);

    auto map = readOctomapTopic();
    auto manager = setupCollisionManager(map, zones);

    for (int i = 0; i < checkpoints.size() - 1; i++)
    {
        plan(checkpoints[i], checkpoints[i+1], map, manager, paths);
    }

    if (paths.size() != checkpoints.size() - 1)
    {
        writeFail();
        return 0;
    }

    writePaths(paths);
    return 1;
}
